/*
 * This code is distributed under an open source license which is commonly
 * called a "BSD" style license.  This license applies only to souce files
 * "dscudkp.h" and "dscudkp.c" in this directory.  The rest of the Diamond
 * Systems Universal Driver ("libdscud5.a") remains proprietary software.
 */

/*
 * Copyright (c) 2003, Diamond Systems Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * * Neither the name of Diamond Systems Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

/*
 * DSCUDKP
 *
 * A module for high speed interrupt processing for all Diamond Systems 
 * boards.  The primary purpose of the module is to read data from the
 * boards as quickly as possible and transfer this this data in larger
 * buffers to usermode.  It also supports user interrupts such as timers.
 *
 * To use this module first you must create the device:
 *
 *     mknod /dev/dscud c 252 0
 *
 * If the device major 252 is in use, you can select any available number.
 * To load the module you run:
 *
 *     insmod dscudkp.o
 *
 * If you selected a major number other than 252 such as 250 you'll need to run
 * insmod with an additional parameter to use it:
 *
 *     insmod dscudkp.o dscud_major=250
 *
 * The module will now be ready to service interrupts.  The rest of the 
 * Universal Driver (libdscud5.a) interacts with it using a method like
 * the code below.  You as the user of this driver do not need to write
 * code like this since it is all internal to the DSCUD API.  It is only 
 * given as an example to help document how this kernel module works.
 *
 *     int fd;
 *     DSCUDKP_CONF kpconf;
 *
 *     fd = open("/dev/dscud", "r");
 *
 *     kpconf.boardtype = DSC_DMM32;  // initialize all of struct
 *     ioctl(fd, IOCTL_DSCUDKP_CONF, &kpconf);
 *
 *     while ( read(fd, mybuf, bufsize) ) {
 *         //process the data;
 *     }
 *
 *     close(fd); 
 *
 * The driver configures the kernel module with board information using the
 * ioctl() call.  See dscudkp.h for information on this struct.
 *
 */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <linux/module.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/time.h>
#include <asm/io.h>

/*
 * Kernel abstraction code taken with permission from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published by O'Reilly &
 * Associates.
 */

#ifndef LINUX_VERSION_CODE
#  include <linux/version.h>
#endif

#ifndef KERNEL_VERSION /* pre-2.1.90 didn't have it */
#  define KERNEL_VERSION(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

/* only allow 2.0.x  2.2.y and 2.4.z */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,0,0) /* not < 2.0 */
#  error "This kernel is too old: not supported by this file"
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) /* not > 2.4, by now */
#  error "This kernel is too recent: not supported by this file"
#endif
#if (LINUX_VERSION_CODE & 0xff00) == 1 /* not 2.1 */
#  error "Please don't use linux-2.1, use 2.2 or 2.4 instead"
#endif
#if (LINUX_VERSION_CODE & 0xff00) == 3 /* not 2.3 */
#  error "Please don't use linux-2.3, use 2.4 instead"
#endif

/* remember about the current version */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
#  define LINUX_20
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#  define LINUX_22
#else
#  define LINUX_24
#endif

#ifdef LINUX_20
#  define EXPORT_NO_SYMBOLS register_symtab(NULL);
#  define MODULE_PARM(v,t)           /* nothing */
#endif

#ifndef SET_MODULE_OWNER
#  define SET_MODULE_OWNER(structure) /* nothing */
#endif

/*
 * Wait queues changed with 2.3
 */
#ifndef DECLARE_WAIT_QUEUE_HEAD
#  define DECLARE_WAIT_QUEUE_HEAD(head) struct wait_queue *head = NULL
   typedef  struct wait_queue *wait_queue_head_t;
#  define init_waitqueue_head(head) (*(head)) = NULL
#endif

#ifdef LINUX_20
#  include <asm/segment.h>
#  define   copy_to_user(t,f,n)         (memcpy_tofs((t), (f), (n)), 0)
#  define copy_from_user(t,f,n)       (memcpy_fromfs((t), (f), (n)), 0)
#else
#  include <asm/uaccess.h>
#endif

#ifdef LINUX_20
#  define __USE_OLD_REBUILD_HEADER__
#endif

#ifdef LINUX_24
#	include <linux/slab.h>
#else
#	include <linux/malloc.h>
#endif

/* END KERNEL ABSTRACTION DEFINITIONS */


#include "dscudkp.h"

/* Timeout when waiting for status bit in AD interrupt scan for dmm+dmm16
 * and DA interrupt for dmm32+hercules */
#define TIMEOUT_MICROSECS 100

/* 
 * Default device major number is 252.  In the unlikely event that this 
 * number is already in use, you can overide this number at kernel module
 * load time like this:
 *
 * bash% insmod dscudkp dscud_major=201
 *
 */
static int dscud_major = 252;
MODULE_PARM(dscud_major, "i");

typedef struct {
	/* Board configuration from DSCUDKP_CONF */
	BYTE boardtype;
	WORD base_address;
	BYTE int_level;
	BYTE flipflop_offset;
	WORD dump_threshold;
	DWORD max_transfers;
	BYTE int_type;
	BYTE high_channel;
	BYTE low_channel;
	BYTE scan_enab;
	BYTE user_int_mode;

	/*
	 * Queue used to wakup read() operations when another dump_threshold
	 * chunk of data is ready to be transfered to usermode.
	 */
	wait_queue_head_t read_queue;

	/*  Flag determines if module will free IRQ on close() */
	int irq_active;

	/* Buffer variables */
	SWORD *samples;  /* dump_threshold or max_transfers sized */
	SWORD *pCurPos;  /* Pointer to current position in *samples */
	DWORD bufsize; /* Size of *samples */

	/* Number of samples to read from fifo.  high_ch - low_ch or fifo depth */
	WORD samples_per_int;

	/* For calulating how often to transfer data to user mode */
	DWORD cur_transfers;

	/* For boards which require software based channel incrementing */
	WORD cur_channel;

	/* Error code if a failure happened.  Uses dscud.h errors */
	BYTE error;

	/* Last DMM48 opto state. Saved in case usermode requests it */
	BYTE dmm48_opto_state;

} KP_DATA;

/*
 * cleanup_module() - Release module resources
 */
void cleanup_module(void)
{
	unregister_chrdev(dscud_major, "dscud");
}

/*
 * dscud_open() - Handle open(/dev/dscud)
 */
static int dscud_open(struct inode *node, struct file *filep)
{
	KP_DATA *kpdata;

	kpdata = (KP_DATA *)vmalloc(sizeof(KP_DATA));
	if ( kpdata == NULL )
		return -ENOMEM;

	memset(kpdata, 0, sizeof(KP_DATA));
	init_waitqueue_head (&(kpdata->read_queue));

	filep->private_data = (void *)kpdata;

	MOD_INC_USE_COUNT; 
	return 0;		
}

/*
 * dscud_release() - Handle close(/dev/dscud).  Free board resources.
 */
static int dscud_release(struct inode *node, struct file *filep)
{
	KP_DATA *kpdata = (KP_DATA *)filep->private_data;

	if ( kpdata->irq_active ) {
		free_irq(kpdata->int_level, kpdata);
	}

	if ( kpdata->samples != NULL ) {
		vfree(kpdata->samples);
		kpdata->samples = NULL;
	}

	vfree(kpdata);
	filep->private_data = NULL;

	MOD_DEC_USE_COUNT;
	return 0;		
}

/*
 * dscud_read() - Handle read(/dev/dscud). Wait for ISR to signal that another
 * buffer is ready for transfer to user mode
 */
static ssize_t dscud_read(struct file *filep, char *buff, size_t count, loff_t *offp)
{
	KP_DATA *kpdata = (KP_DATA *)filep->private_data;
	DWORD expected_bufsize;

	if ( kpdata->int_type == INT_TYPE_AD )
		expected_bufsize = kpdata->bufsize;
	else
		expected_bufsize = 0;

	if ( count != expected_bufsize ) {
		return -EINVAL;
	}

	/* Wait for next chunk of data from the ISR */
	interruptible_sleep_on(&kpdata->read_queue);

	/* Error occured in ISR.  Right now this only happens if there is a
	 * timeout while waiting for a status bit */
	if ( kpdata->error )
		return -(kpdata->error);

	/* ADInt more data available.  Copy it */
	if ( count > 0 ) {
		if ( copy_to_user(buff, kpdata->samples, count) )
			return -EFAULT;
	}

	/* Return to usermode */
	return count;
}

/*
 * busy_wait_status() - implements a function for waiting on a hw status
 * bit to change with a short microsecond timeout.  
 */
int busy_wait_status(int addr, unsigned char reg)
{
	int ticks;

	for ( ticks = 0; (inb(addr) & reg) && ticks < TIMEOUT_MICROSECS; ticks++ )
		udelay(1);

	if ( ticks >= TIMEOUT_MICROSECS ) {
		return ETIME;
	}

	return 0;
}

/* 
 * at_adint() - interrupt processing routine for all FIFO equiped DMM* boards
 * plus the Prometheus and Hercules.  Any Diamond board which fits the basic
 * model of insw(baseaddr, fifo), outb(value, flipflop) can work with this
 * routine.
 */
static void at_adint(int irq, void *dev_id, struct pt_regs *regs)
{
	KP_DATA *kpdata = (KP_DATA *)dev_id;
	int dump = 0;  
	int i;

	/* Ignore interrupt if AINT register not set */
	if ( kpdata->boardtype == DSC_PROM ) {
		if ( !(inb(kpdata->base_address + 7) & 0x10 ) )
			return;
	} else if ( kpdata->boardtype == DSC_DMM48 ) {
		if ( !(inb(kpdata->base_address + 11) & 0x10 ) )
			return;
	} else if ( kpdata->boardtype == DSC_HERCEBX ) {
		if ( !(inb(kpdata->base_address + 14) & 0x10 ) )
			return;
	} else if ( kpdata->boardtype == DSC_DMM16AT || 
					kpdata->boardtype == DSC_DMMAT ) 
	{
		if ( !(inb(kpdata->base_address + 8) & 0x10 ) )
			return;
	} else {
		if ( !(inb(kpdata->base_address + 9) & 0x80 ) ) {
			return;
		}
	}

	/* NOTE: WORKAROUND FOR OLD PRZ BUG
	 * Supress spurious PRZ interrupts when fifo not at correct depth.
	 */
	if ( kpdata->boardtype == DSC_PROM ) {
		BYTE fifo_thresh = inb(kpdata->base_address+5);
		BYTE fifo_depth = inb(kpdata->base_address+6);

		if ( fifo_thresh == 0 )
			fifo_thresh = 1;

		if ( fifo_depth < fifo_thresh ) {
			outb(0x01, kpdata->base_address + kpdata->flipflop_offset);
			return;
		}
	}

	/* Drain the fifo */
	insw(kpdata->base_address, kpdata->pCurPos, kpdata->samples_per_int);

	/* DMMAT only does 12 bit samples */
	if ( kpdata->boardtype == DSC_DMMAT )
		for ( i = 0; i < kpdata->samples_per_int; i++ )
			*(kpdata->pCurPos + i) = (*(kpdata->pCurPos + i) >> 4) & 0x0FFF;

	/* Advance our buffer pointers */
	kpdata->pCurPos += kpdata->samples_per_int;
	kpdata->cur_transfers += kpdata->samples_per_int;

	/* Check to see if now is the time to send the buffer to user mode */
	if ( kpdata->dump_threshold && kpdata->cur_transfers >= kpdata->dump_threshold )
	{
		dump = 1;
	}

	/* Dump time or transfers exceeds the max */
	if ( dump || kpdata->cur_transfers >= kpdata->max_transfers ) {
		kpdata->pCurPos = kpdata->samples;
		kpdata->cur_transfers = 0;
		wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()

	/* Otherwise reset the flipflop and wait for the next interrupt */
	} else {
		if ( kpdata->boardtype == DSC_PROM ) {
			outb(0x01, kpdata->base_address + kpdata->flipflop_offset);
		} else if ( kpdata->boardtype == DSC_DMM48 ) {
			outb(0x10, kpdata->base_address + kpdata->flipflop_offset);
		} else if ( kpdata->boardtype == DSC_HERCEBX ) {
			outb(0x02, kpdata->base_address + kpdata->flipflop_offset);
		} else {
			outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
		}
	}
}


/* 
 * nonat_adint_sample() - Handle sample interrupt for DMM.
 */
static void nonat_adint_sample(int irq, void *dev_id, struct pt_regs *regs)
{
	KP_DATA *kpdata = (KP_DATA *)dev_id;
	int dump = 0;  
	BYTE tmp;

	/* Drain the fifo */
	*kpdata->pCurPos = inw(kpdata->base_address);
	*kpdata->pCurPos = (*kpdata->pCurPos >> 4) & 0x0FFF;

	/* Advance our buffer pointers */
	kpdata->pCurPos++;
	kpdata->cur_transfers++;

	/* Set new channel only if we are sampling on multiple channels */
	if ( kpdata->low_channel != kpdata->high_channel ) {

		kpdata->cur_channel++;
		if ( kpdata->cur_channel > kpdata->high_channel )
			kpdata->cur_channel = kpdata->low_channel;

		outb((BYTE)((kpdata->high_channel << 4) | kpdata->cur_channel) , 
			kpdata->base_address + 2);

		/* Wait for the status bit */
		if ( busy_wait_status(kpdata->base_address + 11, 0x10) ) {
			kpdata->error = ETIME;
			wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
			return;
		}
	}

	/* Check to see if now is the time to send the buffer to user mode */
	if ( kpdata->dump_threshold && kpdata->cur_transfers >= kpdata->dump_threshold )
	{
		dump = 1;
	}

	/* Dump time or transfers exceeds the max */
	if ( dump || kpdata->cur_transfers >= kpdata->max_transfers ) {

		tmp = inb(kpdata->base_address + 9);
		outb((BYTE)(tmp & 0xFD) , kpdata->base_address + 9);

		kpdata->pCurPos = kpdata->samples;
		kpdata->cur_transfers = 0;

		wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()

	/* Otherwise reset the flipflop and wait for the next interrupt */
	} else {
		outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
	}
}


/* 
 * dmm_adint_scan() - Handle scan interrupts for the DMM.  These
 * boards do not directly support scan interrupts so channel selection is
 * driven by software.
 */
static void nonat_adint_scan(int irq, void *dev_id, struct pt_regs *regs)
{
	KP_DATA *kpdata = (KP_DATA *)dev_id;
	int dump = 0;  
	int tmp;
	int ch;

	tmp = inb(kpdata->base_address + 8);

	/* Interrupt sharing not supported for these boards */
	if ( !(tmp & 0x10 ) )
		return;

	/* Disable interrupts */
	tmp = inb(kpdata->base_address + 9);
	outb((BYTE)(tmp & 0x7D) , kpdata->base_address + 9);

	/* Read the first sample */
	*kpdata->pCurPos = inw(kpdata->base_address);
	*kpdata->pCurPos = (*kpdata->pCurPos >> 4) & 0x0FFF;

	/* Advance our buffer pointers */
	kpdata->pCurPos++;
	kpdata->cur_transfers++;

	/* Trigger and read samples for the other channels.  Note the wait
	 * for the status bit, which may cause system problems on hardware failure */
	for ( ch = kpdata->low_channel + 1; ch <= kpdata->high_channel; ch++ ) {

		/* Trigger another conversion */
		outb(0xFF, kpdata->base_address);

		/* Wait for the status bit */
		if ( busy_wait_status(kpdata->base_address + 8, 0x80) ) {
			kpdata->error = ETIME;
			wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
			return;
		}

		/* Read the sample */
		*kpdata->pCurPos = inw(kpdata->base_address);
		*kpdata->pCurPos = (*kpdata->pCurPos >> 4) & 0x0FFF;

		kpdata->pCurPos++;
		kpdata->cur_transfers++;
	}

	/* Reset the channel */
	outb((BYTE)((kpdata->high_channel << 4) | kpdata->low_channel) , 
					kpdata->base_address + 2);

	/* Wait for the status bit */
	if ( busy_wait_status(kpdata->base_address + 11, 0x10) ) {
		kpdata->error = ETIME;
		wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
		return;
	}

	/* Check to see if now is the time to send the buffer to user mode */
	if ( kpdata->dump_threshold && kpdata->cur_transfers >= kpdata->dump_threshold )
	{
		dump = 1;
	}

	/* Dump time or transfers exceeds the max */
	if ( dump || kpdata->cur_transfers >= kpdata->max_transfers ) {

		tmp = inb(kpdata->base_address + 9);
		outb((BYTE)(tmp & 0xFD) , kpdata->base_address + 9);

		kpdata->pCurPos = kpdata->samples;
		kpdata->cur_transfers = 0;
		wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()

	/* Otherwise reset the flipflop and wait for the next interrupt */
	} else {
		tmp = inb(kpdata->base_address + 9);
		outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
		outb((BYTE)(tmp | 0x82) , kpdata->base_address + 9);
	}
}


/* 
 * zmm_adint_sample() - Handle sample interrupts for the ZMM.
 */
static void zmm_adint_sample(int irq, void *dev_id, struct pt_regs *regs)
{
	KP_DATA *kpdata = (KP_DATA *)dev_id;
	int dump = 0;  
	BYTE config;

	/* Interrupt sharing not supported for these boards */

	/* Read the sample */
	*kpdata->pCurPos = inb(kpdata->base_address);

	/* Advance our buffer pointers */
	kpdata->pCurPos += kpdata->samples_per_int;
	kpdata->cur_transfers += kpdata->samples_per_int;

	/* Maintain the current channel variable */
	if ( ++kpdata->cur_channel > kpdata->high_channel )
		kpdata->cur_channel = kpdata->low_channel;

	/* Read config */
	config = inb(kpdata->base_address + 4);

	/* Increment the channel */
	outb(kpdata->cur_channel | config, kpdata->base_address + 4);

	/* Check to see if now is the time to send the buffer to user mode */
	if ( kpdata->dump_threshold && kpdata->cur_transfers >= kpdata->dump_threshold )
	{
		dump = 1;
	}

	/* Dump time or transfers exceeds the max */
	if ( dump || kpdata->cur_transfers >= kpdata->max_transfers ) {
		kpdata->pCurPos = kpdata->samples;
		kpdata->cur_transfers = 0;
		
		wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()

	/* Otherwise reset the flipflop and wait for the next interrupt */
	} else {
		inb(kpdata->base_address + kpdata->flipflop_offset);
	}
}

/* 
 * zmm_adint_scan() - Handle scan interrupts for the ZMM.
 */
static void zmm_adint_scan(int irq, void *dev_id, struct pt_regs *regs)
{
	KP_DATA *kpdata = (KP_DATA *)dev_id;
	int dump = 0;  
	BYTE config;
	int ch;

	/* Interrupt sharing not supported for these boards */

	/* Read the sample */
	*kpdata->pCurPos = inb(kpdata->base_address);

	/* Advance our buffer pointers */
	kpdata->pCurPos += kpdata->samples_per_int;
	kpdata->cur_transfers += kpdata->samples_per_int;

	/* Maintain the current channel variable */
	if ( ++kpdata->cur_channel > kpdata->high_channel )
		kpdata->cur_channel = kpdata->low_channel;

	/* Read config */
	config = inb(kpdata->base_address + 4);

	/* Increment the channel */
	outb(kpdata->cur_channel | config, kpdata->base_address + 4);

	/* Trigger and read samples for the other channels.  Note the wait
	 * for the status bit, which will lock up the system */
	for ( ch = kpdata->low_channel + 1; ch <= kpdata->high_channel; ch++ ) {
		outb(0xFF, kpdata->base_address + 2);   // Trigger another conversion

		/* Wait for the status bit */
		if ( busy_wait_status(kpdata->base_address + 4, 0x80) ) {
			kpdata->error = ETIME;
			wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
			return;
		}

		/* Read the sample */
		*kpdata->pCurPos = inw(kpdata->base_address);
		kpdata->pCurPos++;
		kpdata->cur_transfers++;
	
		/* Maintain the current channel variable */
		if ( ++kpdata->cur_channel > kpdata->high_channel )
			kpdata->cur_channel = kpdata->low_channel;

		outb(kpdata->cur_channel | config, kpdata->base_address + 4);
	}

	/* Check to see if now is the time to send the buffer to user mode */
	if ( kpdata->dump_threshold && kpdata->cur_transfers >= kpdata->dump_threshold )
	{
		dump = 1;
	}

	/* Dump time or transfers exceeds the max */
	if ( dump || kpdata->cur_transfers >= kpdata->max_transfers ) {
		kpdata->pCurPos = kpdata->samples;
		kpdata->cur_transfers = 0;
		
		wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()

	/* Otherwise reset the flipflop and wait for the next interrupt */
	} else {
		inb(kpdata->base_address + kpdata->flipflop_offset);
	}
}


/* 
 * generic_userint() - Handle user interrupts for all boards.  These are timers
 * which require no kernel processing or data transfer.  A routine in usermode
 * will be run.
 */
static void generic_userint(int irq, void *dev_id, struct pt_regs *regs)
{
	KP_DATA *kpdata = (KP_DATA *)dev_id;
	BYTE tmp;

	/* Ignore interrupt if the correct xINT register is not set.
	 * Otherwise reset the flipflop */
	switch (kpdata->boardtype) {
		case DSC_DMM32:
			if ( kpdata->int_type == INT_TYPE_COUNTER 
							|| kpdata->int_type == INT_TYPE_DIOIN ) {
				if ( !(inb(kpdata->base_address + 9) & 0x60) )
					return;
			} else if ( kpdata->int_type == INT_TYPE_AD ) {
				if ( !(inb(kpdata->base_address + 9) & 0x80) )
					return;
			}

			outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
			break;

		case DSC_PROM:
			if ( kpdata->int_type == INT_TYPE_COUNTER || kpdata->int_type == INT_TYPE_USER ) {
				if ( !(inb(kpdata->base_address + 7) & 0x40) )
					return;
				outb(0x04, kpdata->base_address + kpdata->flipflop_offset);
			} else if ( kpdata->int_type == INT_TYPE_DIOIN ) {
				if ( !(inb(kpdata->base_address + 7) & 0x20) )
					return;
				outb(0x02, kpdata->base_address + kpdata->flipflop_offset);

			} else if ( kpdata->int_type == INT_TYPE_AD ) {
				if ( !(inb(kpdata->base_address + 7) & 0x10) )
					return;
				outb(0x01, kpdata->base_address + kpdata->flipflop_offset);
			}
			break;

		case DSC_DMMAT:
		case DSC_DMM16AT:
			if ( kpdata->int_type == INT_TYPE_COUNTER ) {
				if ( !(inb(kpdata->base_address + 8) & 0x40) )
					return;
			} else if ( kpdata->int_type == INT_TYPE_AD ) {
				if ( !(inb(kpdata->base_address + 8) & 0x10) )
					return;
			}
			outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
			break;

		case DSC_DMM:
			tmp = inb(kpdata->base_address + 9);
			outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
			outb((BYTE)(tmp | 0x82) , kpdata->base_address + 9);
			break;

		case DSC_DMM48:
			if ( kpdata->int_type == INT_TYPE_DIOIN ) {
				if ( !(inb(kpdata->base_address + 11) & 0x40) ) 
					return;
				outb(0x40, kpdata->base_address + kpdata->flipflop_offset);
			} else if ( kpdata->int_type == INT_TYPE_COUNTER) {
				if ( !(inb(kpdata->base_address + 11) & 0x80) ) 
					return;
				outb(0x80, kpdata->base_address + kpdata->flipflop_offset);
			} else if ( kpdata->int_type == INT_TYPE_OPTO ) {
				if ( !(inb(kpdata->base_address + 11) & 0x20) ) 
					return;

				kpdata->dmm48_opto_state = inb(kpdata->base_address+7);
				outb(0x20, kpdata->base_address + kpdata->flipflop_offset);
			} else if ( kpdata->int_type == INT_TYPE_AD ) {
				if ( !(inb(kpdata->base_address + 11) & 0x10) ) 
					return;
				outb(0x10, kpdata->base_address + kpdata->flipflop_offset);
			}
			break;

		case DSC_HERCEBX:
			if ( kpdata->int_type == INT_TYPE_COUNTER ) {
				if ( !(inb(kpdata->base_address + 14) & 0x40) ) 
					return;
				outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
			} else if ( kpdata->int_type == INT_TYPE_DIOIN ) {
				if ( !(inb(kpdata->base_address + 14) & 0x20) ) 
					return;
				outb(0x04, kpdata->base_address + kpdata->flipflop_offset);
			} else if ( kpdata->int_type == INT_TYPE_AD ) {
				if ( !(inb(kpdata->base_address + 14) & 0x10) ) 
					return;
				outb(0x02, kpdata->base_address + kpdata->flipflop_offset);
			}
			break;

		case DSC_QMM:
			inb(kpdata->base_address + 7);

	}

	/* Wakes up read() which calls usermode user interrupt function */
	wake_up(&kpdata->read_queue);
}

/*
 * generic_daint() - D/A interrupt support.  See the universal driver
 * user manual for more information.  A buffer is passed from usermode
 * which contains DA codes which are output on a range of channels at
 * a rate specified by the counter or by the AD clock if a counter is
 * not present.
 */
static void generic_daint(int irq, void *dev_id, struct pt_regs *regs)
{
	KP_DATA *kpdata = (KP_DATA *)dev_id;
	BYTE ch, lsb, msb;
	int tmp;

	/* Ignore interrupt if TINT (or INT for DMM/ZMM) isn't set. */
	switch (kpdata->boardtype) {
		case DSC_DMM32:
			if ( !(inb(kpdata->base_address + 9) & 0x20) ) {
				return;
			}
			break;

		case DSC_PROM:
			if ( !(inb(kpdata->base_address + 7) & 0x40) )
				return;
			break;

		case DSC_DMM16AT:
		case DSC_DMMAT:
			if ( !(inb(kpdata->base_address + 8) & 0x40) )
				return;
			break;

		case DSC_HERCEBX:
			if ( !(inb(kpdata->base_address + 14) & 0x40) ) 
				return;
			break;

		case DSC_DMM48:
			if ( !(inb(kpdata->base_address + 11) & 0x80) )
				return;

			break;
	}

	for ( ch = kpdata->low_channel; ch <= kpdata->high_channel; ch++ ) {
		if ( kpdata->boardtype == DSC_DMM32 ) {
			lsb = *kpdata->pCurPos & 0xFF;
			msb = ((*kpdata->pCurPos >> 8) & 0x0F) | (ch<<6);

			outb(lsb, kpdata->base_address + 4);

			/* Wait for the status bit */
			if ( busy_wait_status(kpdata->base_address + 4, 0x80) ) {
				kpdata->error = ETIME;
				wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
				return;
			}

			/* Send the MSB */
			outb(msb, kpdata->base_address + 5);

			/* Wait for the status bit */
			if ( busy_wait_status(kpdata->base_address + 4, 0x80) ) {
				kpdata->error = ETIME;
				wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
				return;
			}

		} else if ( kpdata->boardtype == DSC_DMM16AT ) {
			lsb = *kpdata->pCurPos & 0xFF;
			msb = (*kpdata->pCurPos >> 8) & 0x0F;

			outb(lsb, kpdata->base_address + 1);
			outb(msb, kpdata->base_address + 4 + ch);

		} else if ( kpdata->boardtype == DSC_DMMAT || 
						kpdata->boardtype == DSC_DMM ) 
		{
			outw(*kpdata->pCurPos, kpdata->base_address + 4 + (ch*2));

		} else if ( kpdata->boardtype == DSC_PROM ) {
			lsb = *kpdata->pCurPos & 0xFF;
			msb = ((*kpdata->pCurPos >> 8) & 0x0F) | (ch<<6);

			outb(lsb, kpdata->base_address + 6);
			outb(msb, kpdata->base_address + 7);

			/* Wait for the status bit */
			if ( busy_wait_status(kpdata->base_address + 3, 0x10) ) {
				kpdata->error = ETIME;
				wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
				return;
			}

		} else if ( kpdata->boardtype == DSC_DMM48 ) {
			lsb = *kpdata->pCurPos & 0xFF;
			msb = ((*kpdata->pCurPos >> 8) & 0xFF);

			outb(lsb, kpdata->base_address);
			outb(msb, kpdata->base_address + 1);
			outb(ch, kpdata->base_address + 7);

			/* Wait for the status bit */
			if ( busy_wait_status(kpdata->base_address + 9, 0x40) ) {
				kpdata->error = ETIME;
				wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
				return;
			}

			outb(0x10, kpdata->base_address + 7);

		} else if ( kpdata->boardtype == DSC_HERCEBX ) {
			lsb = *kpdata->pCurPos & 0xFF;
			msb = ((*kpdata->pCurPos >> 8) & 0x0F) | (ch<<6);

			outb(lsb, kpdata->base_address + 6);
			outb(msb, kpdata->base_address + 7);
			outb(ch, kpdata->base_address + 5);

			/* Wait for the status bit */
			if ( busy_wait_status(kpdata->base_address + 4, 0x20) ) {
				kpdata->error = ETIME;
				wake_up_interruptible(&kpdata->read_queue);  // Wakes up read()
				return;
			}

		} else if ( kpdata->boardtype == DSC_ZMM ) {
			msb = (BYTE)*kpdata->pCurPos;
			outb(msb, kpdata->base_address);
		}

		kpdata->pCurPos++;
	}

	/* Triggers the DA conversion */
	if ( kpdata->boardtype == DSC_DMM16AT )
		inb(kpdata->base_address + 4);

	kpdata->cur_transfers += kpdata->samples_per_int;

	if ( kpdata->cur_transfers >= kpdata->max_transfers ) {
		kpdata->cur_transfers = 0;
		kpdata->pCurPos = kpdata->samples;
		wake_up(&kpdata->read_queue);  // Wakes up read()

	} else {

		/* Reset the flipflop */
		switch (kpdata->boardtype) {
			case DSC_DMM32:
			case DSC_DMM16AT:
			case DSC_DMMAT:
			case DSC_HERCEBX:
				outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
				break;

			case DSC_DMM48:
				outb(0x80, kpdata->base_address + kpdata->flipflop_offset);
				break;

			case DSC_PROM:
				outb(0x04, kpdata->base_address + kpdata->flipflop_offset);
				break;

			case DSC_ZMM:
				inb(kpdata->base_address + kpdata->flipflop_offset);
				break;

			case DSC_DMM:
				tmp = inb(kpdata->base_address + 9);
				outb(0x08, kpdata->base_address + kpdata->flipflop_offset);
				outb((BYTE)(tmp | 0x82) , kpdata->base_address + 9);
				break;
		}
	}
}


/*
 * initialize_adint() - Process configuration struct from ioctl() and make
 * sure the configuration is valid.  Register an interrupt handler.
 */
static int initialize_adint(KP_DATA *kpdata, DSCUDKP_CONF *conf)
{
	int rc;

	/* Copy to our kernel plugin specific struct */
	kpdata->boardtype = conf->boardtype;
	kpdata->base_address = conf->base_address;
	kpdata->int_level = conf->int_level;
	kpdata->flipflop_offset = conf->flipflop_offset;
	kpdata->dump_threshold = conf->dump_threshold;
	kpdata->max_transfers = conf->max_transfers;
	kpdata->int_type = conf->int_type;
	kpdata->high_channel = conf->high_channel;
	kpdata->low_channel = conf->low_channel;
	kpdata->scan_enab = conf->scan_enab;
	kpdata->user_int_mode = conf->user_int_mode;

	if ( kpdata->user_int_mode == USER_INT_INSTEAD ) {
		kpdata->samples = NULL;
		rc = request_irq(kpdata->int_level, &generic_userint, 
						SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);
		if (!rc) 
			kpdata->irq_active = 1;

		return rc;
	}

	/* Make sure max_transfers is non zero */
	if ( conf->max_transfers < 1 ) {
		return -EINVAL;
	}

	/* Calculate samples_per_int now to save repetitive caluculations later in
	 * the ISR */
	if ( conf->fifo_enab && conf->fifo_depth > 0 )
		kpdata->samples_per_int = conf->fifo_depth;
	else if ( kpdata->scan_enab )
		kpdata->samples_per_int = 1 + kpdata->high_channel - kpdata->low_channel;
	else
		kpdata->samples_per_int = 1;

	/* Allocate buffer for storing sample data.  This appears to max out at
	 * around 50Kb */
	if ( kpdata->dump_threshold == 0 )
		kpdata->bufsize = sizeof(SWORD) * kpdata->max_transfers;
	else
		kpdata->bufsize = sizeof(SWORD) * kpdata->dump_threshold;

	kpdata->samples = vmalloc(kpdata->bufsize);
	if ( kpdata->samples == NULL )
		return -ENOMEM;

	kpdata->pCurPos = kpdata->samples;
	kpdata->cur_transfers = 0;
	kpdata->irq_active = 1;
	kpdata->cur_channel = kpdata->low_channel;
	rc = -EINVAL;

	/* Register an IRQ handler */
	switch (kpdata->boardtype) {
		case DSC_DMM48:
		case DSC_DMM32:
		case DSC_PROM:
		case DSC_DMM16AT:
		case DSC_DMMAT:
		case DSC_HERCEBX:
			rc = request_irq(kpdata->int_level, &at_adint, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);
			break;

		case DSC_DMM:
			if ( kpdata->scan_enab )
				rc = request_irq(kpdata->int_level, &nonat_adint_scan, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);
			else
				rc = request_irq(kpdata->int_level, &nonat_adint_sample, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);

			break;

		case DSC_ZMM:
			if ( kpdata->scan_enab )
				rc = request_irq(kpdata->int_level, &zmm_adint_scan, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);
			else
				rc = request_irq(kpdata->int_level, &zmm_adint_sample, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);

			break;

		default:
			rc = -EINVAL;
			break;
	}

	/* request_irq() failed or unknown board */
	if ( rc ) {
		vfree(kpdata->samples);
		kpdata->samples = NULL;
		kpdata->pCurPos = NULL;
		kpdata->irq_active = 0;
	}

	return rc;
}

/*
 * initialize_userint() - Process configuration from ioctl() for a user interrupt.
 */
static int initialize_userint(KP_DATA *kpdata, DSCUDKP_CONF *conf)
{
	int rc;

	kpdata->boardtype = conf->boardtype;
	kpdata->base_address = conf->base_address;
	kpdata->int_level = conf->int_level;
	kpdata->flipflop_offset = conf->flipflop_offset;
	kpdata->dump_threshold = conf->dump_threshold;
	kpdata->max_transfers = conf->max_transfers;
	kpdata->int_type = conf->int_type;
	kpdata->high_channel = conf->high_channel;
	kpdata->low_channel = conf->low_channel;
	kpdata->scan_enab = conf->scan_enab;
	kpdata->user_int_mode = conf->user_int_mode;

	rc = request_irq(kpdata->int_level, &generic_userint, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);
	if ( rc )
		return rc;

	kpdata->irq_active = 1;
	kpdata->cur_transfers = 1;
	return 0;
}

/*
 * initialize_daint() - Process configuration from ioctl() for a DA interrupt.
 */
static int initialize_daint(KP_DATA *kpdata, DSCUDKP_CONF *conf)
{
	int rc;

	kpdata->boardtype = conf->boardtype;
	kpdata->base_address = conf->base_address;
	kpdata->int_level = conf->int_level;
	kpdata->flipflop_offset = conf->flipflop_offset;
	kpdata->dump_threshold = conf->dump_threshold;
	kpdata->max_transfers = conf->max_transfers;
	kpdata->int_type = conf->int_type;
	kpdata->high_channel = conf->high_channel;
	kpdata->low_channel = conf->low_channel;
	kpdata->scan_enab = conf->scan_enab;
	kpdata->user_int_mode = conf->user_int_mode;

	kpdata->samples_per_int = 1 + kpdata->high_channel - kpdata->low_channel;
	if ( kpdata->max_transfers % kpdata->samples_per_int )
		return -EINVAL;

	/* Allocate buffer for storing dacode data */
	kpdata->bufsize = sizeof(SWORD) * kpdata->max_transfers;

	kpdata->samples = vmalloc(kpdata->bufsize);
	if ( kpdata->samples == NULL )
		return -ENOMEM;

	kpdata->pCurPos = kpdata->samples;
	kpdata->cur_transfers = 0;
	kpdata->irq_active = 1;
	kpdata->cur_channel = kpdata->low_channel;
	rc = -EINVAL;

	switch (kpdata->boardtype) {
		case DSC_DMM48:
		case DSC_DMM32:
		case DSC_PROM:
		case DSC_DMM16AT:
		case DSC_DMMAT:
		case DSC_HERCEBX:
			rc = request_irq(kpdata->int_level, &generic_daint, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);
			break;

		case DSC_DMM:
			rc = request_irq(kpdata->int_level, &generic_daint, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);
			break;

		case DSC_ZMM:
			rc = request_irq(kpdata->int_level, &generic_daint, SA_INTERRUPT|SA_SHIRQ, "DSCUD", kpdata);
			break;

		default:
			rc = -EINVAL;
			break;
	}

	/* request_irq() failed or unknown board */
	if ( rc ) {
		vfree(kpdata->samples);
		kpdata->samples = NULL;
		kpdata->pCurPos = NULL;
		kpdata->irq_active = 0;
	}

	return rc;
}


/*
 * dscud_ioctl() - Handle ioctl(/dev/dscud) for configuring the board.  See
 * dscudkp.h for information on the configuration struct and the IOCTL 
 * command number.
 */
static int dscud_ioctl(struct inode *node, struct file *filep, unsigned int cmd, 
					unsigned long arg)
{
	KP_DATA *kpdata = (KP_DATA *)filep->private_data;
	DSCUDKP_CONF conf;

	/* ioctl(dev_dscud_fd, IOCTL_DSCUDKP_CONF, &kpconf) */
	if ( cmd == IOCTL_DSCUDKP_CONF ) {

		/* User only allowed to call ioctl successfully once.  If they 
		 * want to reconfigure they most close() and reopen. */
		if ( kpdata->irq_active )
			return -EBUSY;

		/* Copy the *DSCUDKP_CONF from user land to kernel land */
		if ( copy_from_user((void *)&conf, (void *)arg, sizeof(DSCUDKP_CONF)) )
			return -EFAULT;

		switch (conf.int_type) {
			case INT_TYPE_AD:
				return initialize_adint(kpdata, &conf);

			case INT_TYPE_DA:
				return initialize_daint(kpdata, &conf);

			case INT_TYPE_COUNTER:
			case INT_TYPE_DIOIN:
			case INT_TYPE_OPTO:
			case INT_TYPE_USER:
				return initialize_userint(kpdata, &conf);

			default:
				return -EINVAL;
		}

	/* ioctl(dev_dscud_fd, IOCTL_DSCUDKP_DATA, &data) */
	/* For D/A Interrupt support  only */
	} else if ( cmd == IOCTL_DSCUDKP_DATA ) {
		if ( !kpdata->irq_active )
			return -EINVAL;

		if ( kpdata->samples == NULL )
			return -EINVAL;

		if ( copy_from_user((void *)kpdata->samples, (void *)arg, kpdata->bufsize) )
			return -EFAULT;

	} else if ( cmd == IOCTL_DSCUDKP_GET_TRANSFERS ) {

		/* Copy the current transfers to userland */
		if ( copy_to_user((void *)arg, (void *)&kpdata->cur_transfers, sizeof(DWORD)) )
			return -EFAULT;

	} else if ( cmd == IOCTL_DSCUDKP_GET_OPTOSTATE ) {

		/* Copy the last opto state to userland */
		if ( copy_to_user((void *)arg, (void *)&kpdata->dmm48_opto_state, sizeof(BYTE)) )
			return -EFAULT;

	/* ioctl(dev_dscud_fd, UNKNOWN_COMMAND) ? */
	} else {
		return -ENOTTY;
	}	

	return 0;		
}



/*
 * The following wrappers are meant to make things work with 2.0 kernels
 */
#ifdef LINUX_20
int dscud_read_20(struct inode *ino, struct file *f, char *buf, int count)
{
		 return (int)dscud_read(f, buf, count, &f->f_pos);
}

void dscud_release_20(struct inode *ino, struct file *f)
{
		 dscud_release(ino, f);
}

/* Redefine "real" names to the 2.0 ones */
#define dscud_read dscud_read_20
#define dscud_release dscud_release_20
#endif /* LINUX_20 */

/* 
 * Initialize file_operations structure used to attach to events on /dev/dscud 
 * such as open/close.
 */
static struct file_operations dscud_fops = {
	open:  dscud_open,
	release:  dscud_release,
	ioctl:  dscud_ioctl,
	read:  dscud_read,
};


/*
 * init_module() - Initialize the module and load time
 */
int init_module(void)
{
	int rc;

	EXPORT_NO_SYMBOLS;
	SET_MODULE_OWNER(&dscud_fops);

	rc = register_chrdev(dscud_major, "dscud", &dscud_fops);
	if ( rc < 0 )
		return rc;

	return 0;
}


